#include <cds/gc/dhp.h>
#include <cds/gc/hp.h>
#include <cds/init.h>
#include <cds/intrusive/vyukov_mpmc_cycle_queue.h>
#include <iostream>
#include <string>

using namespace std;

namespace ci = cds::intrusive;
typedef cds::gc::HP hp_gc;

struct Foo {
  Foo(int x) : x(x) {}
  int x;
};

typedef cds::intrusive::VyukovMPMCCycleQueue<
    Foo, typename cds::intrusive::vyukov_queue::make_traits<cds::opt::buffer<
             cds::opt::v::uninitialized_static_buffer<Foo, 1024>>>::type>
    static_queue;

// Queue of Foo pointers, capacity is 1024, dynamically allocated buffer:
struct queue_traits : public cds::intrusive::vyukov_queue::traits {
  typedef cds::opt::v::uninitialized_dynamic_buffer<Foo> buffer;
};
typedef cds::intrusive::VyukovMPMCCycleQueue<Foo, queue_traits> dynamic_queue;

int main() {
  cds::Initialize();

  {
    // Initialize Hazard Pointer singleton
    cds::gc::HP hpGC(128, 8, 128);
    // If main thread uses lock-free containers
    // the main thread should be attached to libcds infrastructure
    cds::threading::Manager::attachThread();

    static_queue q;
    Foo *f = new Foo(1);
    Foo *res = nullptr;
    q.enqueue(*f);
    res = q.dequeue();
    if (res) {
      cout << "Dequeued " << res->x << "\n";
    } else {
      cout << "Dequeued none\n";
    }
  }

  cds::Terminate();
  return 0;
}
